import torch
import numpy as np
import matplotlib
import torchvision
import math
import os
import zipfile

from matplotlib.patches import Rectangle
from torch import Tensor
from torch.hub import download_url_to_file
from tqdm import tqdm  # pip3 install tqdm
import matplotlib.pyplot as plt
plt.rcParams["figure.figsize"] = (15, 15)
plt.style.use('dark_background')
import torch.utils.data
import scipy.misc
import scipy.ndimage
import sklearn.decomposition
import argparse # pip3 install argparse
import torch.nn.functional as F
import albumentations as A
import cv2
from yolo.yolo import yolov4, YOLOV4_Weights
from yolo.file_utils import FileUtils
from yolo.data_utils import DataUtils


BATCH_SIZE = 2
USE_CUDA = torch.cuda.is_available()
DEVICE = 'cpu'
if USE_CUDA:
    DEVICE = 'cuda'


class DatasetLLUWoodKnots(torch.utils.data.Dataset):
    def __init__(self, part):

        self.part = part
        dataset_name = f'llu_wood_knots_{part}.npy'
        path_dataset = f'../data/{dataset_name}'
        path_bounding_boxes = path_dataset.replace(f'llu_wood_knots_{part}.npy', f'llu_wood_knots_bounding_boxes_{part}.json')
        if not os.path.exists(path_dataset):
            FileUtils.createDir('../data')
            download_url_to_file(
                f'http://share.yellowrobot.xyz/1651675667-llu/{dataset_name}',
                path_dataset,
                progress=True
            )

            download_url_to_file(
                f'http://share.yellowrobot.xyz/1651675667-llu/llu_wood_knots_bounding_boxes_{part}.json',
                path_bounding_boxes,
                progress=True
            )

        self.np_data = np.load(path_dataset, allow_pickle=False)
        # (RGBK, W, H)
        # K channel means Knot label

        self.list_bounding_boxes = FileUtils.loadJSON(path_bounding_boxes)
        self.len_max_bounding_boxes_length = max([len(bounding_boxes) for bounding_boxes in self.list_bounding_boxes])

        self.augment_transformations = A.Compose([
            A.HorizontalFlip(p=0.5),
            A.VerticalFlip(p=0.5),
            A.Affine(
                scale=(0.8, 1.2),
                rotate=(-90, 90),
                p=0.9,
                mode=cv2.BORDER_REFLECT
            ),
            A.ColorJitter(
                brightness=0.2,
                contrast=0.2,
                saturation=0.2,
                hue=0.2,
                p=0.5
            ),
        ])

    def __len__(self):
        return len(self.np_data)

    def __getitem__(self, idx):
        np_rgbk = self.np_data[idx] # red, green, blue, knots
        bounding_boxes = self.list_bounding_boxes[idx]

        debug = 'none'
        if self.part == 'train':
            if np.random.random() < 0.5:
                debug = 'augment'
                result = self.augment_transformations(image=np_rgbk[:3, :, :].transpose(1, 2, 0), mask=np_rgbk[3, :, :])
                np_rgbk[:3, :, :] = result['image'].transpose(2, 0, 1)
                np_rgbk[3, :, :] = result['mask']

                # recalculate bounding boxes
                bounding_boxes = DataUtils.get_bboxes_from_image(np_rgbk[3])
                if len(bounding_boxes) == 0:
                    # if aufgmentation failed, use the original image
                    np_rgbk = self.np_data[idx]
                    bounding_boxes = self.list_bounding_boxes[idx]

        bbox_len = len(bounding_boxes)
        bbox = torch.zeros((self.len_max_bounding_boxes_length, 4)).float()
        if len(bounding_boxes) > self.len_max_bounding_boxes_length:
            bounding_boxes = bounding_boxes[:self.len_max_bounding_boxes_length]
        bbox[:bbox_len, :] = torch.from_numpy(np.array(bounding_boxes))

        # plt.cla()
        # plt.clf()
        # plt.imshow(np_rgbk[0:3].transpose(1, 2, 0)/255)
        # for bbox_each in bounding_boxes:
        #     if bbox_each[2] > 0:  # ignore empty boxes
        #         plt.gca().add_patch(Rectangle(
        #             (bbox_each[0], bbox_each[1]),
        #             bbox_each[2] - bbox_each[0],
        #             bbox_each[3] - bbox_each[1],
        #             linewidth=1, edgecolor='r', facecolor='none'
        #         ))
        # #plt.show()
        # plt.savefig(f'./tmp/{debug}_{idx}.png')

        x = np_rgbk[0:3] / 255.0
        y = np_rgbk[3]

        x = torch.FloatTensor(x)
        y = torch.FloatTensor(y[np.newaxis, :])

        return x, y, bbox, bbox_len

dataset_train = DatasetLLUWoodKnots(part='train')
dataset_test = DatasetLLUWoodKnots(part='test')

data_loader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=True
)
data_loader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=BATCH_SIZE,
    shuffle=False,
    drop_last=True
)

model = yolov4() # can also load pretrained weights from Darknet

optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)

if USE_CUDA:
    model = model.cuda()

metrics = {}
metrics_epoch = {}
for stage in ['train', 'test']:
    for metric in [
        'loss',
        'classification',
        'confidence',
        'overlap',
        'dice_coef',
        'iou_box',
        'classification',
        'bbox_regression',
        'bbox_ctrness',
        'acc'
    ]:
        metrics[f'{stage}_{metric}'] = []

    for metric in [
        'loss_best',
        'dice_coef_best',
        'iou_box_best',
        'acc_best'
    ]:
        metrics[f'{stage}_{metric}'] = float('-inf') if metric != 'loss_best' else float('inf')

for epoch in range(1, 100):
    metrics_epoch = {key: [] for key in metrics.keys()}

    for data_loader in [data_loader_train, data_loader_test]:
        for x, y, bbox, bbox_len in tqdm(data_loader, desc=f'{stage} epoch {epoch}'):
            if data_loader == data_loader_test:
                stage = 'test'
                model = model.eval()
                torch.set_grad_enabled(False)
            else:
                stage = 'train'
                model = model.train()
                torch.set_grad_enabled(True)

            x = x.to(DEVICE)
            y = y.to(DEVICE).squeeze(dim=1)
            bbox = bbox.to(DEVICE)

            bbox_dicts = []

            # TODO implement model training

            loss: Tensor = 0
            for key in out.keys():
                loss += out[key]
                metrics_epoch[f'{stage}_{key}'].append(out[key].item())
            metrics_epoch[f'{stage}_loss'].append(loss.item())

            if data_loader == data_loader_train:
                loss.backward()
                optimizer.step()

            optimizer.zero_grad()

            # TODO implement model inference

            # TODO filter out the predictions that overlap with lower confidence scored boxes
            # TODO boxes already ordered by confidence score
            # TODO filter also out zero labelled boxes as model uses softmax with 2 classes


            # TODO  torchvision.ops.box_iou(boxes1: torch.Tensor, boxes2: torch.Tensor)
            # TODO  for each ground truth bbox find highest iou for bbox_prim then calculate average iou for whole batch

            metrics_epoch[f'{stage}_acc'].append(np.mean(batch_acc))
            batch_io = np.mean(batch_ios)
            metrics_epoch[f'{stage}_iou_box'].append(batch_io)

            loss = loss.cpu()
            x = x.cpu()
            y = y.cpu()
            bbox = bbox.cpu()
            bbox_len = bbox_len.cpu()
            bbox_prim = bbox_prim.cpu()

            np_x = x.data.numpy()
            np_y = y.data.numpy()

        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                if '_best' not in key:
                    value = np.mean(metrics_epoch[key])
                    metrics_epoch[key] = value
                    metrics[key].append(value)
                    metrics_strs.append(f'{key}: {round(value, 2)}')

                    if key + '_best' in metrics.keys():
                        if 'loss' in key:
                            metrics[key + '_best'] = min(metrics[key][-1], metrics[key + '_best'])
                        else:
                            metrics[key + '_best'] = max(metrics[key][-1], metrics[key + '_best'])
                        metrics_epoch[key + '_best'] = metrics[key + '_best']

        print(f'epoch: {epoch} {" ".join(metrics_strs)}')


    #plt.subplot(121) # row col idx
    plt.clf()
    plt.cla()
    plt.subplot(321)  # row col idx
    plt.plot(metrics['train_loss'], label='train_loss')
    plt.plot(metrics['test_loss'], label='test_loss')
    plt.legend()

    plt.subplot(323)  # row col idx
    plt.plot(metrics['train_acc'], label='train_acc')
    plt.plot(metrics['test_acc'], label='test_acc')
    plt.legend()

    plt.subplot(325)  # row col idx
    plt.plot(metrics['train_iou_box'], label='train_iou_box')
    plt.plot(metrics['test_iou_box'], label='test_iou_box')
    plt.legend()

    np_x = np_x.transpose(0, 2, 3, 1)
    np_x = np_x * 255
    np_x = np_x.astype(np.uint8)

    for i, j in enumerate([4, 5, 6, 16, 17, 18, 10, 11, 12, 22, 23, 24]):
        if i < len(np_x):
            plt.subplot(4, 6, j)
            plt.title('Blue - True, Red - Predicted')
            plt.imshow(np_x[i], interpolation=None)
            plt.imshow(np_y[i], cmap='Blues', alpha=0.5, interpolation=None)

            bboxes = bbox[i, :].numpy() # ground truth boxes
            for bbox_each in bboxes:
                if bbox_each[2] > 0: # ignore empty boxes
                    plt.gca().add_patch(Rectangle(
                        (bbox_each[0], bbox_each[1]),
                        bbox_each[2] - bbox_each[0],
                        bbox_each[3] - bbox_each[1],
                        linewidth=1, edgecolor='b', facecolor='none'
                    ))

            bboxes = bbox_prim[i, :].numpy() # predicted boxes
            for bbox_each in bboxes:
                if bbox_each[2] > 0: # ignore empty boxes
                    plt.gca().add_patch(Rectangle(
                        (bbox_each[0], bbox_each[1]),
                        bbox_each[2] - bbox_each[0],
                        bbox_each[3] - bbox_each[1],
                        linewidth=1, edgecolor='r', facecolor='none'
                    ))

    plt.tight_layout(pad=0)
    plt.show()
